% FFT based method for calculating the error surface for the epipole
% (focus of expansion, FOE) for motion from calibrated flow.
%
% Input
%
% foewin = [exmin,exmax; eymin,eymax] window of FOE positions to
%	calculate, in the same coordinates as image pixels.
%
% fx,fy = input images containing x,y components of optical flow
%
% K = camera calibration matrix, e.g. [fx 0 copx; 0 fy copy; 0 0 1].
%
% fweight = image of relative weightings for different image points,
%	default 1.
%
% fcov = shared covariance matrix of fx,fy measurements, default eye(2).
%
% Output
%
% foeerr = (exmax-exmin+1)*(eymax-eymin+1) image containing FOE error surface
%
% est = [err,foex,foey,wx,wy,wz] best candidate for FOE and angular
%	velocity w. WARNING: for exactly planar scenes there are two
%	zero-error minima, and the wrong one may be chosen even for
%	slightly non-planar scenes.
%
% Conventions
%
% x is the first (fastest, matrix row) index, y the second (slowest,
% matrix column) one. The equations (sign of angular velocity w) assume
% the usual x right y down (i.e. left handed) image coordinates --
% adding z forward gives right handed 3D ones).
%
% Reference
%
% S. Srinivasan, Extracting Structure from Optical Flow using the Fast
% Error Search Technique, International Journal of Computer Vision,
% 37:3, June 2000. (http://www.cfar.umd.edu/~shridhar/ - this also has
% some C code, but I haven't tried it).
%

function [foeerr,est] = srinivasan(foewin,fx,fy,K,fweight,fcov)
   epsilon2 = 1e-2;
   if nargin<6
      fcov = [];
      if nargin<5, fweight = 1, end;
   end
   IK = inv(K);

   % Set up sizes for image (mi,ni) and foe window (me,ne). To avoid
   % wrap-around effects, we have to FFT zero-padded matrices of size
   % (mt,nt)=(mi,ni)+(me,ne). (We don't round these up to a power of
   % two for now, but could).

   exmin = foewin(1,1); exmax = foewin(1,2);
   eymin = foewin(2,1); eymax = foewin(2,2);
   me = exmax-exmin+1;
   ne = eymax-eymin+1;
   [mi,ni] = size(fx);
   mt = mi + me;
   nt = ni + ne;

   % Fill the foe-sensitivity-field matrices. Entries S__(di,dj) give
   % the reduced information (contribution weights) for the pixel
   % (ex-exmin,ey-eymin) + (di,dj) to an FOE at (ex,ey) (i.e. entry
   % (ex-exmin+1,ey-eymin+1) of the output matrix).  This holds directly
   % for (1,1)<=(di,dj)<=(mi,ni). For mi<di<=mt and/or ni<dj<=nt, the
   % entries hold the weights for negative offsets -me<di-mt<=0 and
   % -ne<dj-nt<=0. (NB: (di,dj)=(1,1) corresponds to "zero shift": for
   % an FOE window identical to the image, the (pixel)-(FOE) shift is
   % (di-1,dj-1) not (di,dj)).

   dx = ([0:me-1,-mi:-1] + exmin-1)'*ones(1,nt);
   dy = ones(mt,1)*([0:ne-1,-ni:-1] + eymin-1);

   dxc = IK(1,1)*dx + IK(1,2)*dy;
   dyc = IK(2,1)*dx + IK(2,2)*dy;
   if (size(fcov,2)<2)
      finfo = K(1:2,1:2)'*K(1:2,1:2);
   else
      finfo = K(1:2,1:2)'*inv(fcov)*K(1:2,1:2);
   end

   % Find calibrated-coordinate epipolar-line-orthogonal projected flow
   % information matrix S.  Adding epsilon2 avoids numerical problems by
   % reducing the contribution of pixels very close to the FOE. More
   % generally, we could modulate the sensitivity field by a
   % position-dependent weighting if we wanted, e.g. to damp
   % contributions near to or far from the epipole.

   Sx = finfo(1,1)*dxc + finfo(1,2)*dyc;
   Sy = finfo(2,1)*dxc + finfo(2,2)*dyc;
   S2 = Sx.*dxc + Sy.*dyc;
   Snorm = S2 + epsilon2;
   Sxx = (finfo(1,1)*S2 - Sx.^2)./Snorm;
   Sxy = (finfo(1,2)*S2 - Sx.*Sy)./Snorm;
   Syy = (finfo(2,2)*S2 - Sy.^2)./Snorm;

   % Old misguided version of damping. This destroys the error surface
   % as it includes some translational flow contribution for points near
   % the FOE...
   %
   %    if (size(fcov,2)<2)
   %       % Default flow covariance = eye(2).
   %       Snorm = (dx.^2 + dy.^2 + epsilon2);
   %       Sxx = 1 - (dx.^2)./Snorm;
   %       Sxy =   - (dx.*dy)./Snorm;
   %       Syy = 1 - (dy.^2)./Snorm;
   %    else
   %       % Nonuniform flow covariance.
   %       finfo = inv(fcov);
   %       Sx = finfo(1,1)*dx + finfo(1,2)*dy;
   %       Sy = finfo(2,1)*dx + finfo(2,2)*dy;
   %       Snorm = (Sx.*dx + Sy.*dy + epsilon2);
   %       Sxx = finfo(1,1) - (Sx.^2)./Snorm;
   %       Sxy = finfo(1,2) - (Sx.*Sy)./Snorm;
   %       Syy = finfo(2,2) - (Sy.^2)./Snorm;
   %    end

   % FFT the sensitivity fields to prepare for convolution.  We avoid a
   % conjugation here by using backwards intervals in dx,dy above.

   Wxx = fft2(Sxx); Wxy = fft2(Sxy); Wyy = fft2(Syy);

   % Now work out on the image (x rather than e-x) terms.  Matrices
   % Bx,By,Fx,Fy contain the calibrated coordinates and flow of pixel
   % (i,j). I.e. (Bx,By,1) is the bona fide 3D direction vector in the
   % camera frame of the points imaged in pixel (i,j).

   ix = [1:mi]' * ones(1,ni);
   iy = ones(mi,1) * [1:ni];
   Bx = IK(1,1)*ix + IK(1,2)*iy + IK(1,3);
   By = IK(2,1)*ix + IK(2,2)*iy + IK(2,3);
   Fx = IK(1,1)*fx + IK(1,2)*fy;
   Fy = IK(2,1)*fx + IK(2,2)*fy;

   % From the calibrated motion geometry Bx,By and flow Fx,Fy, we build
   % the matrix B: for a given angular velocity w, B*[w;1] is the
   % estimated translational flow (rotational_flow(w) -
   % measured_flow). If the estimate is correct, we can null the
   % translational flow by projecting orthogonal to epipolar lines. This
   % ultimately gives the cost function, which is a sum (over tensor
   % indices) of convolutions of the above FOE sensitivities S (which
   % project orthogonal to epipolar lines) and terms from B2, the tensor
   % product of B with itself, which we can write as a 3 x 10 set of
   % fields.

   % B:    X    Y   Z   F    
   %    -------------------
   %  x | -xy  X1  -y  -u |       (X1,Y1) = (1+x^2,1+y^2)
   %  y | -Y1  xy   x  -v |       (u,v) = (fx,fy)
   %    -------------------

   % B2:   (XX       XY)      (YY       XZ)   (YZ      ZZ)   (XF      YF)   (ZF   FF)
   %    -------------------------------------------------------------------------------
   % xx | (xy)^2   -xyX1      X1^2     xy^2   -X1y     y^2   xyu     -X1u    yu   u^2 |
   % xy | 2xyY1 -X1Y1-(xy)^2 2xyX1  yY1-x^2y X1x-xy^2 -2xy xyv+Y1u -X1v-xyu yv-xu 2uv |
   % yy |  Y1^2    -xyY1     (xy)^2   -xY1     x^2y    x^2   Y1v     -xyv    -xv  v^2 |
   %    -------------------------------------------------------------------------------

   % We group these in pairs and use real and imaginary parts to halve
   % the number of FFT's needed. Any pairing based on columns suffices
   % for this as the only complexity in FFT space is by rows: we must
   % contract over xx,xy,yy indices after convolution (multiplication)
   % in FFT space, then IFFT the 5 results (one real/imaginary pair per
   % 2 columns) and unpack the real and imaginary parts to find the
   % resulting 10 real columns.  The pairing shown separates the 4 flow
   % columns which may be the only data-dependent ones.
   %
   % Also note that B_y_Y = -B_x_X = xy implies that 3 of the 3*10=30
   % entries in B2 are duplicated:
   %
   % B2_yy_YY = B2_xx_XX,   B2_xy_YY = -2*B2_xx_XY,   B2_xy_XX = -2*B2_yy_XY
   %
   % So in principle only 27 entries need to be evaluated. It might be
   % possible to achieve this using real transforms (e.g. Hartley)
   % instead of complex FFT's the convolutions, but with FFT's I don't
   % see how to do it: grouping as above puts the duplicated
   % xx_(XX+i*YY) components in one group, but we can't unpick the
   % results in Fourier space even though we have both of the
   % transformed components from other vectors.

   Bxx = Bx.^2;
   Bxy = Bx.*By;
   Byy = By.^2;
   BX1 = 1+Bxx;
   BY1 = 1+Byy;

   % (XX,XY) component
   %
   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((Bxy.^2) - i*(Bxy.*BX1));		% xx 
   A = Wxx.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((2*Bxy.*BY1) - i*(BX1.*BY1+Bxy.^2)); % xy
   A = A + Wxy.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((BY1.^2) - i*(Bxy.*BY1));		% yy
   A = A + Wyy.*fft2(M);
   B = ifft2(A);
   B = B(1:me,1:ne);
   MXX = real(B);
   MXY = imag(B);

   % (YY,XZ) component
   %
   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((BX1.^2) + i*(Bx.*Byy));		% xx 
   A = Wxx.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((2*Bxy.*BX1) + i*((BY1-Bxx).*By));	% xy
   A = A + Wxy.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((Bxy.^2) - i*(Bx.*BY1));		% yy
   A = A + Wyy.*fft2(M);
   B = ifft2(A);
   B = B(1:me,1:ne);
   MYY = real(B);
   MXZ = imag(B);

   % (YZ,ZZ) component
   %
   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*(-(BX1.*By) + i*Byy);		% xx 
   A = Wxx.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*(((BX1-Byy).*Bx) - (2*i)*Bxy);	% xy
   A = A + Wxy.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*(Bxx.*(By + i));			% yy
   A = A + Wyy.*fft2(M);
   B = ifft2(A);
   B = B(1:me,1:ne);
   MYZ = real(B);
   MZZ = imag(B);

   % (XF,YF) component
   %
   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((Bxy - i*BX1).*Fx);			% xx 
   A = Wxx.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((Bxy-i*BX1).*Fy+(BY1-i*Bxy).*Fx);	% xy
   A = A + Wxy.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((BY1-i*Bxy).*Fy);			% yy
   A = A + Wyy.*fft2(M);
   B = ifft2(A);
   B = B(1:me,1:ne);
   MXF = real(B);
   MYF = imag(B);

   % (ZF,FF) component
   %
   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((By+i*Fx).*Fx);			% xx 
   A = Wxx.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((By+(2*i)*Fx).*Fy - Bx.*Fx);	% xy
   A = A + Wxy.*fft2(M);

   M = zeros(mt,nt);
   M(1:mi,1:ni) = fweight.*((-Bx+i*Fy).*Fy);			% yy
   A = A + Wyy.*fft2(M);
   B = ifft2(A);
   B = B(1:me,1:ne);
   MZF = real(B);
   MFF = imag(B);

   % We now have an me x ne image of symmetric 4x4 matrices M expressed
   % as 10 matrix-component-images (MXX...MFF).  Each M=M(FOE)
   % determines a quadratic cost [w,1]*M*[w;1] where w is the angular
   % velocity. Minimizing over w gives the minimal cost for each FOE:
   % cost(FOE) = MFF - b'*inv(A)*b where b=[MXF;MYF;MZF] and A is the
   % 3x3 MXX..MZZ matrix.  The simplest way to find this cost is by
   % unpivoted Cholesky or LDL' decomposition of M, the (4,4) element
   % being the cost. Rather than doing me x ne calls to chol(), we do it
   % by hand for all points at once.

   if 0
      % Freeze w=0, for testing
      foeerr = MFF;
   else
      % Solution for free w. This is M = chol(M) for each FOE position.

      MXX = sqrt(MXX);
      MXY = MXY./MXX;
      MXZ = MXZ./MXX;
      MXF = MXF./MXX;
      MYY = sqrt(MYY-MXY.^2);
      MYZ = (MYZ-MXY.*MXZ)./MYY;
      MYF = (MYF-MXY.*MXF)./MYY;
      MZZ = sqrt(MZZ-MXZ.^2-MYZ.^2);
      MZF = (MZF-MXZ.*MXF-MYZ.*MYF)./MZZ;
      foeerr = (MFF-MXF.^2-MYF.^2-MZF.^2);
      % MFF = sqrt(foeerr);
   end

   % Find the minimum over the error matrix, and back solve for the
   % minimizing rotation.

   c = min(min(foeerr));
   [x0,y0] = find(foeerr == c);
   x0 = x0(1); y0 = y0(1);

   % A = [MXX(x0,y0),MXY(x0,y0),MXZ(x0,y0); ...
   %   	  0,MYY(x0,y0),MYZ(x0,y0); ...
   %   	  0,0,MZZ(x0,y0)];
   %    b = [MXF(x0,y0); MYF(x0,y0); MZF(x0,y0)];
   % w = -A \ b

   wz = -MZF(x0,y0)/MZZ(x0,y0);
   wy = -(MYF(x0,y0)+MYZ(x0,y0)*wz)/MYY(x0,y0);
   wx = -(MXF(x0,y0)+MXY(x0,y0)*wy+MXZ(x0,y0)*wz)/MXX(x0,y0);
   
   est = [c; x0+exmin-1; y0+eymin-1; wx; wy; wz];

   if 0
      % Here is how to back solve for the minimizing rotation
      % w=-inv(A)*b at each FOE position in parallel.
   
      wz = -MZF./MZZ;
      wy = -(MYF+MYZ.*wz)./MYY;
      wx = -(MXF+MXY.*wy+MXZ.*wz)./MXX;
   end

%end;
